# Scola version 1.1.0
# ScolaParser.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Models
# 4) Blocks
# 5) Scenarios
# 6) Instructions
# 7) Expressions
# 8) Auxiliary methods

# 1) Imported modules
# -------------------

import string
import sys
import re
from ScolaLexer import *
from ScolaModels import *

# 2) Creator
# ----------

class ScolaParser:
  MAXIMUM = 100000
  _Functors = {
    ScolaToken.KEYWORD_AND: (ScolaExpression.AND, 2, MAXIMUM),
    ScolaToken.KEYWORD_OR: (ScolaExpression.OR, 2, MAXIMUM),
    ScolaToken.KEYWORD_NOT: (ScolaExpression.NOT, 1, 1),
    ScolaToken.KEYWORD_EQ: (ScolaExpression.EQ, 2, 2),
    ScolaToken.KEYWORD_DF: (ScolaExpression.DF, 2, 2),
    ScolaToken.KEYWORD_LT: (ScolaExpression.LT, 2, 2),
    ScolaToken.KEYWORD_GT: (ScolaExpression.GT, 2, 2),
    ScolaToken.KEYWORD_LEQ: (ScolaExpression.LEQ, 2, 2),
    ScolaToken.KEYWORD_GEQ: (ScolaExpression.GEQ, 2, 2),
    ScolaToken.KEYWORD_OPP: (ScolaExpression.OPP, 1, 1),
    ScolaToken.KEYWORD_INV: (ScolaExpression.INV, 1, 1),
    ScolaToken.KEYWORD_ADD: (ScolaExpression.ADD, 2, MAXIMUM),
    ScolaToken.KEYWORD_SUB: (ScolaExpression.SUB, 2, MAXIMUM),
    ScolaToken.KEYWORD_MUL: (ScolaExpression.MUL, 2, MAXIMUM),
    ScolaToken.KEYWORD_DIV: (ScolaExpression.DIV, 2, MAXIMUM),
    ScolaToken.KEYWORD_MIN: (ScolaExpression.MIN, 1, MAXIMUM),
    ScolaToken.KEYWORD_MAX: (ScolaExpression.MAX, 1, MAXIMUM),
    ScolaToken.KEYWORD_ABS: (ScolaExpression.ABS, 1, 1),
    ScolaToken.KEYWORD_POW: (ScolaExpression.POW, 2, 2),
    ScolaToken.KEYWORD_EXP: (ScolaExpression.EXP, 1, 1),
    ScolaToken.KEYWORD_LOG: (ScolaExpression.LOG, 1, 1),
    ScolaToken.KEYWORD_SQRT: (ScolaExpression.SQRT, 1, 1),
    ScolaToken.KEYWORD_MOD: (ScolaExpression.MOD, 2, 2),
    ScolaToken.KEYWORD_CEIL: (ScolaExpression.CEIL, 1, 1),
    ScolaToken.KEYWORD_FLOOR: (ScolaExpression.FLOOR, 1, 1),
    ScolaToken.KEYWORD_COUNT: (ScolaExpression.COUNT, 1, MAXIMUM),
    ScolaToken.KEYWORD_APPEND: (ScolaExpression.APPEND, 2, MAXIMUM),
    ScolaToken.KEYWORD_IF: (ScolaExpression.IF, 3, 3),
    ScolaToken.KEYWORD_INTEGER: (ScolaExpression.CAST_INTEGER, 1, 1),
    ScolaToken.KEYWORD_REAL: (ScolaExpression.CAST_REAL, 1, 1),
    ScolaToken.KEYWORD_STRING: (ScolaExpression.CAST_STRING, 1, 1),
    ScolaToken.KEYWORD_SYMBOL: (ScolaExpression.CAST_SYMBOL, 0, 1),
    ScolaToken.KEYWORD_IDENTIFIER: (ScolaExpression.IDENTIFIER, 1, 1),
    ScolaToken.KEYWORD_OWNER: (ScolaExpression.OWNER, 1, 1),
    ScolaToken.KEYWORD_IS_BLOCK: (ScolaExpression.IS_BLOCK, 1, 1),
    ScolaToken.KEYWORD_IS_PORT: (ScolaExpression.IS_PORT, 1, 1),
    ScolaToken.KEYWORD_IS_ASSERTION: (ScolaExpression.IS_ASSERTION, 1, 1),
    ScolaToken.KEYWORD_ELEMENT: (ScolaExpression.ELEMENT, 2, 2),
    ScolaToken.KEYWORD_SIZE: (ScolaExpression.SIZE, 1, 1),
    ScolaToken.KEYWORD_QUOTE: (ScolaExpression.QUOTE, 1, 1),
    ScolaToken.KEYWORD_EVAL: (ScolaExpression.EVAL, 1, 1)
    }

  def __init__(self):
    self._lexer = ScolaLexer()
    self._errors = 0
    self._errorFile = sys.stderr
    self._currentTokenIndex = 0
    self._model = None

# 3) Models
# ---------

  def GetNumberOfErrors(self):
    return self._errors

  def GetLexer(self):
    return self._lexer

  def GetModel(self):
    return self._model
  def SetModel(self, model):
    self._model = model

  def ParseFile(self, fileName, file):
    self._errors = self._lexer.Initialize(fileName, file)
    if self._errors:
      return
    self._ParseModel()
    self._lexer.Finalize()
    return self._errors

  def _ParseModel(self):
    self._currentTokenIndex = 0
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None:
        break
      elif token.type==ScolaToken.KEYWORD_DOMAIN:
        self._ParseDomainDeclaration(self._model)
      elif token.type==ScolaToken.KEYWORD_CLONES:
        self._ParseClonesDirective(self._model)
      elif token.type==ScolaToken.KEYWORD_CLASS:
        self._ParseClassDeclaration(self._model)
      elif token.type==ScolaToken.KEYWORD_BLOCK:
        self._ParseBlockDeclaration(self._model)
      elif token.type==ScolaToken.KEYWORD_SCENARIO:
        self._ParseScenarioDeclaration(self._model)
      else:
        self._RaiseParseError(token, "A declaration of domain, class, clone, block or scenario was expected ")
        self._currentTokenIndex += 1
    self._model.SetChecked(False)

  def _ParseDomainDeclaration(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_DOMAIN)
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    domain = ScolaDomain(identifier, container)
    enumeration = ScolaBaseType.NewDomain(identifier)
    domain.SetDomain(enumeration)
    token = self._ParseGivenToken(ScolaToken.LBRACE)
    if token==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    while True:
      token = self._ParseGivenToken(ScolaToken.IDENTIFIER)
      if token==None:
        self._RecoverError(ScolaToken.KEYWORD_END)
        return None
      enumeration.AddSymbolicConstant(token.string)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.type!=ScolaToken.COMMA:
        break
      token = self._ParseGivenToken(ScolaToken.COMMA)
    token = self._ParseGivenToken(ScolaToken.RBRACE)
    if token==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    while True:
      token0 = self._lexer.GetToken(self._currentTokenIndex)
      if token0==None or token0.type==ScolaToken.KEYWORD_END:
        break
      token0 = self._ParseGivenToken(ScolaToken.KEYWORD_LT)
      if token0==None:
        self._RecoverError(ScolaToken.KEYWORD_END)
        return None
      token = self._ParseGivenToken(ScolaToken.IDENTIFIER)
      if token==None:
        self._RecoverError(ScolaToken.KEYWORD_END)
        return None
      constant1 = ScolaConstant.NewSymbol(token.string)
      token = self._ParseGivenToken(ScolaToken.IDENTIFIER)
      if token==None:
        self._RecoverError(ScolaToken.KEYWORD_END)
        return None
      constant2 = ScolaConstant.NewSymbol(token.string)
      ok = enumeration.AddOrderConstraint(constant1, constant2)
      if not ok:
        self._RaiseParseError(token0, "Invalid order relation")
        self._RecoverError(ScolaToken.KEYWORD_END)
        return None
      leastMembers = enumeration.GetLeastMembers()
      if len(leastMembers)!=1:
        self._RaiseParseError(token, "Invalid order relation")
        return None
    token = self._ParseGivenToken(ScolaToken.KEYWORD_END)
    if token==None:
      return None
    return domain

  def _ParseClassDeclaration(self, container):
    token0 = self._ParseGivenToken(ScolaToken.KEYWORD_CLASS)
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    token = self._ParseGivenToken(ScolaToken.LPAR)
    if token==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      self._RaiseParseError(token0, "Unexpected end of file")
      return None
    if token.type==ScolaToken.KEYWORD_BLOCK:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_BLOCK)
      eventualType = ScolaModel.BLOCK
      classDeclaration = ScolaClass(identifier, container)
      classDeclaration.SetEventualType(eventualType)
    elif token.type==ScolaToken.KEYWORD_SCENARIO:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_SCENARIO)
      eventualType = ScolaModel.SCENARIO
      classDeclaration = ScolaClass(identifier, container)
      classDeclaration.SetEventualType(eventualType)
    else:
      identifierOfInheritedClass = self._ParseIdentifier()
      if identifierOfInheritedClass==None:
        return None
      element = self._model.LookForNamedElement(identifierOfInheritedClass)
      if element==None:
        self._RaiseParseError(token0, "Attempt to inherit an undeclared class (%s)" % identifierOfInheritedClass)
        return None
      if element.GetType()!=ScolaModel.CLASS:
        self._RaiseParseError(token0, "%s is not a class" % identifierOfInheritedClass)
        return None
      inheritedClass = element
      eventualType = inheritedClass.GetEventualType()
      classDeclaration = ScolaClass(identifier, container)
      classDeclaration.SetIdentifierOfInheritedClass(identifierOfInheritedClass)
      classDeclaration.SetInheritedClass(inheritedClass)
      classDeclaration.SetEventualType(eventualType)
    token = self._ParseGivenToken(ScolaToken.RPAR)
    if eventualType==ScolaModel.BLOCK: 
      return self._ParseBlockBody(classDeclaration)
    if eventualType==ScolaModel.SCENARIO:
      return self._ParseScenarioBody(classDeclaration)
    return None

  def _ParseClonesDirective(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_CLONES)
    pathToClonedElement = self._ParsePath()
    if pathToClonedElement==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    tokenAs = self._ParseGivenToken(ScolaToken.KEYWORD_AS)
    if tokenAs==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    absolutePathToContainer = container.GetAbsolutePath()
    absolutePathToClonedElement = ScolaPath.Merge(absolutePathToContainer, pathToClonedElement)
    clonedElement = self._model.AccessNamedElement(absolutePathToClonedElement)
    if clonedElement==None:
      self._RaiseParseError(token, "Attempt to clone an undeclared element")
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    if clonedElement.GetType()==ScolaModel.CLONES:
      eventualType = clonedElement.GetEventualType()
    elif clonedElement.GetType()==ScolaModel.INSTANCE:
      eventualType = clonedElement.GetEventualType()
    elif clonedElement.GetType()==ScolaModel.BLOCK:
      eventualType = ScolaModel.BLOCK
    elif clonedElement.GetType()==ScolaModel.SCENARIO:
      eventualType = ScolaModel.SCENARIO
    else:
      self._RaiseParseError(token, "Element %s cannot be cloned" % pathToClonedElement.ToString())
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    clonesDirective = ScolaClonesDirective(identifier, container)
    clonesDirective.SetEventualType(eventualType)
    clonesDirective.SetPathToClonedElement(pathToClonedElement)
    clonesDirective.SetClonedElement(clonedElement)
    if eventualType==ScolaModel.BLOCK:
      return self._ParseBlockBody(clonesDirective)
    if eventualType==ScolaModel.SCENARIO:
      self._ParseScenarioAsDirective(clonesDirective)
      return self._ParseScenarioBody(clonesDirective)
    return None

  def _ParseInheritsDirective(self, container):
    token0 = self._ParseGivenToken(ScolaToken.KEYWORD_INHERITS)
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      self._RaiseParseError(token0, "Unexpected end of file")
      return None
    if token.type==ScolaToken.KEYWORD_BLOCK:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_BLOCK)
      return ScolaInherits(ScolaModel.BLOCK, container)
    if token.type==ScolaToken.KEYWORD_SCENARIO:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_SCENARIO)
      return ScolaInherits(ScolaModel.SCENARIO, container)
    pathToInheritedElement = self._ParsePath()
    if pathToInheritedElement==None:
      return None
    absolutePathToInheritedElement = ScolaPath.Merge(container.GetAbsolutePath(), pathToInheritedElement)
    inheritedElement = self._model.AccessNamedElement(absolutePathToInheritedElement)
    if inheritedElement==None:
      self._RaiseParseError(token0, "Attempt to inherit an undeclared element (%s)" % absolutePathToInheritedElement.ToString())
      return None
    if inheritedElement.GetType()==ScolaModel.CLASS:
      eventualType = inheritedElement.GetEventualType()
    elif inheritedElement.GetType()==ScolaModel.INSTANCE:
      eventualType = inheritedElement.GetEventualType()
    elif inheritedElement.GetType()==ScolaModel.CLONES:
      eventualType = inheritedElement.GetEventualType()
    elif inheritedElement.GetType()==ScolaModel.BLOCK:
      eventualType = ScolaModel.BLOCK
    elif inheritedElement.GetType()==ScolaModel.SCENARIO:
      eventualType = ScolaModel.SCENARIO
    else:
      self._RaiseParseError(token0, "Attempt to inherit an element that cannot be inherited (%s)" % absolutePathToInheritedElement.ToString())
      return None
    if inheritedType!=container.GetType():
      self._RaiseParseError(token0, "Attempt to inherit an element that cannot be inherited (%s)" % absolutePathToInheritedElement.ToString())
      return None
    inheritsDirective = ScolaInherits(inheritedType, container)
    inheritsDirective.SetPathToInheritedElement(pathToInheritedElement)
    inheritsDirective.SetInheritedElement(inheritedElement)
    return directive

# 4) Blocks
# ---------

  def _ParseBlockDeclaration(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_BLOCK)
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    block = ScolaBlock(identifier, container)
    return self._ParseBlockBody(block)

  def _ParseBlockBody(self, container):
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.type==ScolaToken.KEYWORD_END:
        break
      blockField = self._ParseBlockField(container)
      if blockField==None:
        self._RecoverError(ScolaToken.KEYWORD_END)
        return None
    token = self._ParseGivenToken(ScolaToken.KEYWORD_END)
    if token==None:
      return None
    return container

  def _ParseBlockField(self, container):
    field = None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      return field
    if token.type==ScolaToken.KEYWORD_CLONES:
      field = self._ParseClonesDirective(container)
    elif token.type==ScolaToken.KEYWORD_INHERITS:
      field = self._ParseInheritsDirective(container)
    elif token.type==ScolaToken.KEYWORD_PORT:
      field = self._ParsePortDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_BOOLEAN:
      field = self._ParsePortDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_INTEGER:
      field = self._ParsePortDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_REAL:
      field = self._ParsePortDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_SYMBOL:
      field = self._ParsePortDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_STRING:
      field = self._ParsePortDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_BLOCK:
      field = self._ParseBlockDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_ASSERTION:
      field = self._ParseAssertionDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_SET:
      field = self._ParseSetInstruction(container)
    elif token.type==ScolaToken.IDENTIFIER:
      identifier = token.string
      element = self._model.LookForNamedElement(identifier)
      if element==None:
        self._RaiseParseError(token, "Undeclared type %s" % identifier)
        return None
      elif element.GetType()==ScolaModel.DOMAIN:
        field = self._ParsePortDeclaration(container)
      elif element.GetType()==ScolaModel.CLASS:
        field = self._ParseBlockInstanceDeclaration(container)
      else:
        self._RaiseParseError(token, "Unknown type")
        return None
    else:
      self._RaiseParseError(token, "A declaration of port or block was expected")
    return field

  def _ParseBlockInstanceDeclaration(self, container):
    token = self._lexer.GetToken(self._currentTokenIndex)
    classIdentifier = self._ParseIdentifier()
    if classIdentifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    classDeclaration = self._model.LookForNamedElement(classIdentifier)
    if classDeclaration==None:
      self._RaiseParseError(token, "No class %s" % classIdentifier)
      return None
    if classDeclaration.GetType()!=ScolaModel.CLASS:
      self._RaiseParseError(token, "%s is not a class" % classIdentifier)
      return None
    if classDeclaration.GetEventualType()!=ScolaModel.BLOCK:
      self._RaiseParseError(token, "%s is not a class describing a block" % classIdentifier)
      return None
    instance = ScolaInstance(identifier, container)
    instance.SetClassIdentifier(classIdentifier)
    instance.SetClassDeclaration(classDeclaration)
    instance.SetEventualType(classDeclaration.GetEventualType())
    return self._ParseBlockBody(instance)

  def _ParsePortDeclaration(self, container):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      return None
    if token.type==ScolaToken.KEYWORD_PORT:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_PORT)
      baseType = ScolaBaseType.NewAny()
    elif token.type==ScolaToken.KEYWORD_BOOLEAN:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_BOOLEAN)
      baseType = ScolaBaseType.NewBoolean()
    elif token.type==ScolaToken.KEYWORD_INTEGER:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_INTEGER)
      baseType = ScolaBaseType.NewInteger()
    elif token.type==ScolaToken.KEYWORD_REAL:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_REAL)
      baseType = ScolaBaseType.NewReal()
    elif token.type==ScolaToken.KEYWORD_SYMBOL:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_SYMBOL)
      baseType = ScolaBaseType.NewSymbol()
    elif token.type==ScolaToken.KEYWORD_STRING:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_STRING)
      baseType = ScolaBaseType.NewString()
    elif token.type==ScolaToken.IDENTIFIER:
      token = self._ParseGivenToken(ScolaToken.IDENTIFIER)
      path = ScolaPath.FromString(token.string)
      element = self._model.AccessNamedElement(path)
      if element==None:
        self._RaiseParseError(token, "Unknown type")
        return None
      elif element.GetType()==ScolaModel.DOMAIN:
        baseType = element.GetDomain()
      else:
        self._RaiseParseError(token, "Unknown type")
        return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      return None
    expression = self._ParseExpression()
    if expression==None:
      return None
    port = ScolaPort(baseType, identifier, expression, container)
    return port

  def _ParseAssertionDeclaration(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_ASSERTION)
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    assertion = ScolaAssertion(identifier, container)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.type==ScolaToken.KEYWORD_END:
        break
      assertionInstruction = self._ParseInstruction(False, assertion)
      if assertionInstruction==None:
        self._RecoverError(ScolaToken.KEYWORD_END)
        return None
    token = self._ParseGivenToken(ScolaToken.KEYWORD_END)
    if token==None:
      return None
    return assertion

# 5) Scenarios
# ------------

  def _ParseScenarioDeclaration(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_SCENARIO)
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    scenario = ScolaScenario(identifier, container)
    self._ParseScenarioAsDirective(scenario)
    return self._ParseScenarioBody(scenario)

  def _ParseScenarioAsDirective(self, container):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.type!=ScolaToken.KEYWORD_AS:
      return
    token = self._ParseGivenToken(ScolaToken.KEYWORD_AS)
    path = self._ParsePath()
    if path==None:
      return None
    container.SetAsPath(path)

  def _ParseScenarioBody(self, container):
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.type==ScolaToken.KEYWORD_END:
        break
      scenarioField = self._ParseScenarioField(container)
      if scenarioField==None:
        self._RecoverError(ScolaToken.KEYWORD_END)
        return None
    token = self._ParseGivenToken(ScolaToken.KEYWORD_END)
    if token==None:
      return None
    return container

  def _ParseScenarioField(self, container):
    field = None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      return field
    if token.type==ScolaToken.KEYWORD_CLONES:
      field = self._ParseClonesDirective(container)
    elif token.type==ScolaToken.KEYWORD_INHERITS:
      field = self._ParseInheritsDirective(container)
    elif token.type==ScolaToken.KEYWORD_STATE:
      field = self._ParseStateDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_TASK:
      field = self._ParseTaskDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_TEST:
      field = self._ParseTestGatewayDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_CHOICE:
      field = self._ParseChoiceGatewayDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_FORK:
      field = self._ParseForkGatewayDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_JOIN:
      field = self._ParseJoinGatewayDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_SPLIT:
      field = self._ParseSplitGatewayDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_MERGE:
      field = self._ParseMergeGatewayDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_MEET:
      field = self._ParseMeetGatewayDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_SCENARIO:
      field = self._ParseScenarioDeclaration(container)
    elif token.type==ScolaToken.KEYWORD_NEXT:
      field = self._ParseNextDirective(container)
    elif token.type==ScolaToken.IDENTIFIER:
      path = ScolaPath.FromString(token.string)
      element = self._model.AccessNamedElement(path)
      if element==None:
        self._RaiseParseError(token, "Unknown type")
        return None
      elif element.GetType()==ScolaModel.CLASS:
        field = self._ParseScenarioInstanceDeclaration(container)
      else:
        self._RaiseParseError(token, "Unknown type")
        return None
    else:
      self._RaiseParseError(token, "A scenario field was expected")
    return field

  def _ParseScenarioInstanceDeclaration(self, container):
    classIdentifier = self._ParseIdentifier()
    if classIdentifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    classDeclaration = self._model.LookForNamedElement(classIdentifier)
    if classDeclaration==None:
      self._RaiseParseError(token, "No class %s" % classIdentifier)
      return None
    if classDeclaration.GetType()!=ScolaModel.CLASS:
      self._RaiseParseError(token, "%s is not a class" % classIdentifier)
      return None
    if classDeclaration.GetEventualType()!=ScolaModel.SCENARIO:
      self._RaiseParseError(token, "%s is not a class describing a scenario" % classIdentifier)
      return None
    instance = ScolaInstance(identifier, container)
    instance.SetClassIdentifier(classIdentifier)
    instance.SetClassDeclaration(classDeclaration)
    instance.SetEventualType(classDeclaration.GetEventualType())
    self._ParseScenarioAsDirective(instance)
    return self._ParseScenarioBody(instance)

  def _ParseStateDeclaration(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_STATE)
    identifier = self._ParseIdentifier()
    if identifier==None:
      return None
    state = ScolaState(identifier, container)
    return state

  def _ParseTaskDeclaration(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_TASK)
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    task = ScolaTask(identifier, container)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.type==ScolaToken.KEYWORD_END:
        break
      instruction = self._ParseInstruction(True, task)
      if instruction==None:
        self._RecoverError(ScolaToken.KEYWORD_END)
        return None
    token = self._ParseGivenToken(ScolaToken.KEYWORD_END)
    if token==None:
      return None
    return task

  def _ParseTestGatewayDeclaration(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_TEST)
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    gateway = ScolaTestGateway(identifier, container)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.type==ScolaToken.KEYWORD_END:
        break
      declaration = self._ParseGatewayCaseDeclaration(gateway)
      if declaration==None:
        self._RecoverError(ScolaToken.KEYWORD_END)
        return None
    token = self._ParseGivenToken(ScolaToken.KEYWORD_END)
    if token==None:
      return None
    return gateway

  def _ParseGatewayCaseDeclaration(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_CASE)
    identifier = self._ParseIdentifier()
    if identifier==None:
      return None
    condition = self._ParseExpression()
    if condition==None:
      return None
    declaration = ScolaGatewayCase(identifier, condition, container)
    return declaration

  def _ParseChoiceGatewayDeclaration(self, container):
    return self._ParseBranchedGatewayDeclaration(ScolaToken.KEYWORD_CHOICE, container)

  def _ParseBranchedGatewayDeclaration(self, keyword, container):
    token = self._ParseGivenToken(keyword)
    identifier = self._ParseIdentifier()
    if identifier==None:
      self._RecoverError(ScolaToken.KEYWORD_END)
      return None
    if keyword==ScolaToken.KEYWORD_CHOICE:
      gateway = ScolaChoiceGateway(identifier, container)
    elif keyword==ScolaToken.KEYWORD_FORK:
      gateway = ScolaForkGateway(identifier, container)
    elif keyword==ScolaToken.KEYWORD_JOIN:
      gateway = ScolaJoinGateway(identifier, container)
    elif keyword==ScolaToken.KEYWORD_SPLIT:
      gateway = ScolaSplitGateway(identifier, container)
    elif keyword==ScolaToken.KEYWORD_MERGE:
      gateway = ScolaMergeGateway(identifier, container)
    elif keyword==ScolaToken.KEYWORD_MEET:
      gateway = ScolaMeetGateway(identifier, container)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.type==ScolaToken.KEYWORD_END:
        break
      declaration = self._ParseBranchDeclaration(gateway)
      if declaration==None:
        self._RecoverError(ScolaToken.KEYWORD_END)
        return None
    token = self._ParseGivenToken(ScolaToken.KEYWORD_END)
    if token==None:
      return None
    return gateway

  def _ParseBranchDeclaration(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_BRANCH)
    identifier = self._ParseIdentifier()
    if identifier==None:
      return None
    declaration = ScolaGatewayBranch(identifier, container)
    return declaration

  def _ParseForkGatewayDeclaration(self, container):
    return self._ParseBranchedGatewayDeclaration(ScolaToken.KEYWORD_FORK, container)

  def _ParseJoinGatewayDeclaration(self, container):
    return self._ParseBranchedGatewayDeclaration(ScolaToken.KEYWORD_JOIN, container)

  def _ParseSplitGatewayDeclaration(self, container):
    return self._ParseBranchedGatewayDeclaration(ScolaToken.KEYWORD_SPLIT, container)

  def _ParseMergeGatewayDeclaration(self, container):
    return self._ParseBranchedGatewayDeclaration(ScolaToken.KEYWORD_MERGE, container)

  def _ParseMeetGatewayDeclaration(self, container):
    return self._ParseBranchedGatewayDeclaration(ScolaToken.KEYWORD_MEET, container)

  def _ParseNextDirective(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_NEXT)
    sourcePath = self._ParsePath()
    if sourcePath==None:
      return None
    targetPath = self._ParsePath()
    if targetPath==None:
      return None
    connection = ScolaNextDirective(sourcePath, targetPath, container)
    return connection

# 6) Instructions
# ---------------

  def _ParseInstruction(self, inTask, container):
    instruction = None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      return instruction
    if token.type==ScolaToken.KEYWORD_FAIL:
      instruction = self._ParseFailInstruction(container)
    elif token.type==ScolaToken.KEYWORD_SET:
      instruction = self._ParseSetInstruction(container)
    elif token.type==ScolaToken.KEYWORD_NEW and inTask:
      instruction = self._ParseNewInstruction(container)
    elif token.type==ScolaToken.KEYWORD_DELETE and inTask:
      instruction = self._ParseDeleteInstruction(container)
    elif token.type==ScolaToken.KEYWORD_MOVE and inTask:
      instruction = self._ParseMoveInstruction(container)
    elif token.type==ScolaToken.KEYWORD_SEND and inTask:
      instruction = self._ParseSendInstruction(container)
    elif token.type==ScolaToken.KEYWORD_RECEIVE and inTask:
      instruction = self._ParseReceiveInstruction(container)
    elif token.type==ScolaToken.KEYWORD_CLONE and inTask:
      instruction = self._ParseCloneInstruction(container)
    elif token.type==ScolaToken.KEYWORD_IF:
      instruction = self._ParseIfInstruction(inTask, container)
    elif token.type==ScolaToken.KEYWORD_BEGIN:
      instruction = self._ParseInstructionBlock(inTask, container)
    elif token.type==ScolaToken.KEYWORD_PRINT and inTask:
      instruction = self._ParsePrintInstruction(container)
    else:
      self._RaiseParseError(token, "An instruction was expected")
    return instruction

  def _ParseFailInstruction(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_FAIL)
    instruction = ScolaFailInstruction(container)
    return instruction

  def _ParseSetInstruction(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_SET)
    pathExpression = self._ParseExpression()
    if pathExpression==None:
      return None
    valueExpression = self._ParseExpression()
    if valueExpression==None:
      return None
    instruction = ScolaSetInstruction(pathExpression, valueExpression, container)
    return instruction

  def _ParseNewInstruction(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_NEW)
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      self._RaiseParseError(token, "An type was expected")
      return None
    if token.type==ScolaToken.KEYWORD_PORT:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_PORT)
      baseType = ScolaBaseType.NewAny()
      isPort = True
    elif token.type==ScolaToken.KEYWORD_NONE:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_NONE)
      baseType = ScolaBaseType.NewNone()
      isPort = True
    elif token.type==ScolaToken.KEYWORD_BOOLEAN:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_BOOLEAN)
      baseType = ScolaBaseType.NewBoolean()
      isPort = True
    elif token.type==ScolaToken.KEYWORD_INTEGER:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_INTEGER)
      baseType = ScolaBaseType.NewInteger()
      isPort = True
    elif token.type==ScolaToken.KEYWORD_REAL:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_REAL)
      baseType = ScolaBaseType.NewReal()
      isPort = True
    elif token.type==ScolaToken.KEYWORD_SYMBOL:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_SYMBOL)
      baseType = ScolaBaseType.NewSymbol()
      isPort = True
    elif token.type==ScolaToken.KEYWORD_STRING:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_STRING)
      baseType = ScolaBaseType.NewString()
      isPort = True
    elif token.type==ScolaToken.KEYWORD_BLOCK:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_BLOCK)
      blockClass = None
      isPort = False
    elif token.type==ScolaToken.IDENTIFIER:
      token = self._ParseGivenToken(ScolaToken.IDENTIFIER)
      element = self._model.LookForNamedElement(token.string)
      if element==None:
        self._RaiseParseError(token, "An type was expected")
        return None
      if element.GetType()==ScolaModel.DOMAIN:
        baseType = element.GetDomain()
        isPort = True
      elif element.GetType()==ScolaModel.CLASS:
        eventualType = element.GetEventualType()
        if eventualType==ScolaModel.BLOCK:
          blockClass = element
          isPort = False
      else:
        self._RaiseParseError(token, "A block class was expected")
        return None
    else:
      self._RaiseParseError(token, "A type was expected")
      return None
    if isPort:
      pathExpression = self._ParseExpression()
      if pathExpression==None:
        return None
      initializationExpression = self._ParseExpression()
      if initializationExpression==None:
        return None
      instruction = ScolaNewPortInstruction(baseType, pathExpression, initializationExpression, container)
      return instruction
    pathExpression = self._ParseExpression()
    if pathExpression==None:
      return None
    instruction = ScolaNewBlockInstruction(blockClass, pathExpression, container)
    return instruction

  def _ParseDeleteInstruction(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_DELETE)
    pathExpression = self._ParseExpression()
    if pathExpression==None:
      return None
    instruction = ScolaDeleteInstruction(pathExpression, container)
    return instruction

  def _ParseMoveInstruction(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_MOVE)
    sourcePathExpression = self._ParseExpression()
    if sourcePathExpression==None:
      return None
    targetPathExpression = self._ParseExpression()
    if targetPathExpression==None:
      return None
    instruction = ScolaMoveInstruction(sourcePathExpression, targetPathExpression, container)
    return instruction

  def _ParseSendInstruction(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_SEND)
    sourcePathExpression = self._ParseExpression()
    if sourcePathExpression==None:
      return None
    receiverPathExpression = self._ParseExpression()
    if receiverPathExpression==None:
      return None
    instruction = ScolaSendInstruction(sourcePathExpression, receiverPathExpression, container)
    return instruction

  def _ParseReceiveInstruction(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_RECEIVE)
    sourcePathExpression = self._ParseExpression()
    if sourcePathExpression==None:
      return None
    senderPathExpression = self._ParseExpression()
    if senderPathExpression==None:
      return None
    targetPathExpression = self._ParseExpression()
    if targetPathExpression==None:
      return None
    instruction = ScolaReceiveInstruction(sourcePathExpression, senderPathExpression, targetPathExpression, container)
    return instruction

  def _ParseCloneInstruction(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_CLONE)
    sourcePathExpression = self._ParseExpression()
    if sourcePathExpression==None:
      return None
    targetPathExpression = self._ParseExpression()
    if targetPathExpression==None:
      return None
    instruction = ScolaCloneInstruction(sourcePathExpression, targetPathExpression, container)
    return instruction

  def _ParseIfInstruction(self, inTask, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_IF)
    condition = self._ParseExpression()
    if condition==None:
      return None
    token = self._ParseGivenToken(ScolaToken.KEYWORD_THEN)
    if token==None:
      return None
    instruction = ScolaIfInstruction(condition, container)
    thenInstruction = self._ParseInstruction(inTask, instruction)
    if thenInstruction==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token!=None and token.type==ScolaToken.KEYWORD_ELSE:
      token = self._ParseGivenToken(ScolaToken.KEYWORD_ELSE)
      elseInstruction = self._ParseInstruction(inTask, instruction)
      if elseInstruction==None:
        return None
    return instruction

  def _ParseInstructionBlock(self, inTask, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_BEGIN)
    instructionBlock = ScolaInstructionBlock(container)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.type==ScolaToken.KEYWORD_END:
        break
      instruction = self._ParseInstruction(inTask, instructionBlock)
      if instruction==None:
        self._RecoverError(ScolaToken.KEYWORD_END)
        return None
    token = self._ParseGivenToken(ScolaToken.KEYWORD_END)
    if token==None:
      return None
    return instructionBlock

  def _ParsePrintInstruction(self, container):
    token = self._ParseGivenToken(ScolaToken.KEYWORD_PRINT)
    expression = self._ParseExpression()
    if expression==None:
      return None
    instruction = ScolaPrintInstruction(expression, container)
    return instruction

# 7) Expressions
# --------------

  def _ParseExpression(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      return None
    if token.type==ScolaToken.KEYWORD_NONE:
      self._currentTokenIndex += 1
      return ScolaExpression(ScolaExpression.NONE)
    if token.type==ScolaToken.KEYWORD_TRUE:
      self._currentTokenIndex += 1
      return ScolaExpression(ScolaExpression.TRUE)
    if token.type==ScolaToken.KEYWORD_FALSE:
      self._currentTokenIndex += 1
      return ScolaExpression(ScolaExpression.FALSE)
    if token.type==ScolaToken.INTEGER:
      self._currentTokenIndex += 1
      return ScolaExpression.NewInteger(int(token.string))
    if token.type==ScolaToken.REAL:
      self._currentTokenIndex += 1
      return ScolaExpression.NewReal(float(token.string))
    if token.type==ScolaToken.IDENTIFIER:
      self._currentTokenIndex += 1
      return ScolaExpression.NewSymbol(token.string)
    if token.type==ScolaToken.PATH:
      self._currentTokenIndex += 1
      return ScolaExpression.NewReference(token.string)
    if token.type==ScolaToken.STRING:
      self._currentTokenIndex += 1
      return ScolaExpression.NewString(token.string)
    if token.type==ScolaToken.LPAR:
      return self._ParseParenthesedExpression()
    if token.type==ScolaToken.QUOTE:
      return self._ParseQuotedExpression()
    self._RaiseParseError(token, "An expression was expected")
    return None

  def _ParseParenthesedExpression(self):
    token = self._ParseGivenToken(ScolaToken.LPAR)
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      self._RaiseParseError(token, "Unexpected end of line")
      return None
    (functor, minimumNumberOfArguments, maximumNumberOfArguments) = ScolaParser._Functors.get(token.type, (None, None, None))
    if functor==None:
      self._RaiseParseError(token, "A functor was expected")
      self._RecoverError(ScolaToken.RPAR)
      return None
    self._currentTokenIndex += 1
    expression = ScolaExpression(functor)
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or token.type==ScolaToken.RPAR:
        break
      childExpression = self._ParseExpression()
      if childExpression==None:
        self._RecoverError(ScolaToken.RPAR)
        return None
      expression.AddChild(childExpression)
    token = self._ParseGivenToken(ScolaToken.RPAR)
    if token==None:
      return None
    if expression.GetNumberOfChildren()<minimumNumberOfArguments:
      self._RaiseParseError(token, "Too few arguments")
      return None
    if expression.GetNumberOfChildren()>maximumNumberOfArguments:
      self._RaiseParseError(token, "Too many arguments")
      return None
    return expression

  def _ParseQuotedExpression(self):
    token = self._ParseGivenToken(ScolaToken.QUOTE)
    childExpression = self._ParseExpression()
    if childExpression==None:
      return None
    expression = ScolaExpression(ScolaExpression.QUOTE)
    expression.AddChild(childExpression)
    return expression

# 8) Auxiliary methods
# --------------------

  def _ParseIdentifier(self):
    token = self._ParseGivenToken(ScolaToken.IDENTIFIER)
    if token==None:
      return None
    return token.string

  def _ParsePath(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None or (token.type!=ScolaToken.IDENTIFIER and token.type!=ScolaToken.PATH):
      self._RaiseParseError(token, "A path was expected")
      return None
    self._currentTokenIndex += 1
    return ScolaPath.FromString(token.string)

  def _ParseGivenToken(self, tokenType):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None or token.type!=tokenType:
      self._RaiseParseError(token, "A %s was expected" % ScolaToken.GetNameOfType(tokenType))
      return None
    self._currentTokenIndex += 1
    return token

  def _RecoverError(self, tokenType):
    currentTokenIndex = self._currentTokenIndex
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None:
        break
      self._currentTokenIndex += 1
      if token.type==tokenType:
        break
    if currentTokenIndex==self._currentTokenIndex:
      sys.exit()

  def GetErrorFile(self):
    return self._errorFile
  def SetErrorFile(self, file):
    self._errorFile = file

  def _RaiseParseError(self, token, message):
    self._errors += 1
    if token!=None:
      self._errorFile.write("Parse error line %d (%d):\n" % (token.lineNumber, token.columnNumber))
    self._errorFile.write("%s\n\n" % message)